// This file is part of AllMyToes.
//
// AllMyToes is free software: you can redistribute it and/or modify it under the terms
// of the GNU General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// AllMyToes is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
// without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with AllMyToes.
// If not, see <https://www.gnu.org/licenses/>.
//
mod errors;
mod provider;
mod request;
mod thumb;
mod thumbsize;

mod mimesniff;
mod source;
mod sourcedescriptor;
mod writer;

use crate::provider::ProviderList;
use crate::request::ThumbRequest;

use log::{debug, trace};
use once_cell::unsync::OnceCell;
use std::env;
use std::path::{Path, PathBuf};

pub use errors::ToeErrorType;
pub use thumb::Thumb;
pub use thumbsize::ThumbSize;

#[derive(Debug, Clone)]
pub struct AMTConfiguration {
    pub force_creation: bool,
    pub return_smallest_feasible: bool,
    pub leak: bool,
    pub custom_provider_spec_file: Option<String>,
    pub force_inbuilt_provider_spec: bool,
}

impl Default for AMTConfiguration {
    fn default() -> Self {
        Self {
            force_creation: false,
            return_smallest_feasible: true,
            leak: false,
            custom_provider_spec_file: None,
            force_inbuilt_provider_spec: false,
        }
    }
}

#[derive(Debug)]
pub struct AMT {
    configuration: AMTConfiguration,
    provider_list: OnceCell<Result<ProviderList, ToeErrorType>>,
}

fn get_config_directory() -> Result<PathBuf, ToeErrorType> {
    let mut cache_dir: PathBuf = match env::var("XDG_CONFIG_HOME") {
        Ok(v) => {
            trace!(
                "Using config directory from XDG_CONFIG_HOME env variable: {}",
                v
            );
            v.as_str().into()
        }
        Err(_) => match env::var("HOME") {
            Ok(v) => {
                trace!("XDG_CONFIG_HOME env variable not set. Using ~/.config.");
                [v.as_str(), ".config"].iter().collect()
            }
            Err(e) => return Err(ToeErrorType::XDGConfigDirNotIdentifiable(format!("{}", e))),
        },
    };
    cache_dir.push("allmytoes");
    Ok(cache_dir)
}

fn get_provider_spec_file(
    configuration: &AMTConfiguration,
) -> Result<Option<PathBuf>, ToeErrorType> {
    if let Some(provider_file) = &configuration.custom_provider_spec_file {
        trace!(
            "Looking for specific provider config file at {}.",
            provider_file
        );
        let provider_file_path = PathBuf::from(&provider_file);
        if provider_file_path.is_file() {
            Ok(Some(provider_file_path))
        } else {
            Err(ToeErrorType::ProviderSpecFileDoesNotExist(String::from(
                provider_file,
            )))
        }
    } else {
        let mut user_file = get_config_directory()?;
        user_file.push("provider.yaml");
        trace!(
            "Looking for provider config file in XDG configuration directory: {:?}",
            user_file
        );
        if user_file.exists() {
            trace!("Found provider config file in XDG configuration directory.");
            Ok(Some(user_file))
        } else {
            trace!("Did not find provider config file in default location.");
            Ok(None)
        }
    }
}

fn load_provider_list(configuration: &AMTConfiguration) -> Result<ProviderList, ToeErrorType> {
    if configuration.force_inbuilt_provider_spec {
        debug!("Loading default (built in) provider configuration (forced)");
        ProviderList::default()
    } else if let Some(provider_config_file_path) = get_provider_spec_file(configuration)? {
        debug!(
            "Loading provider configuration from {:?}",
            provider_config_file_path
        );
        ProviderList::load(provider_config_file_path)
    } else {
        debug!("Loading default (built in) provider configuration");
        ProviderList::default()
    }
}

impl AMT {
    pub fn new(configuration: &AMTConfiguration) -> Self {
        AMT {
            configuration: configuration.clone(),
            provider_list: OnceCell::new(),
        }
    }

    pub fn get(&self, path: &Path, size: ThumbSize) -> Result<Thumb, ToeErrorType> {
        ThumbRequest {
            path,
            size,
            amt: self,
        }
        .get()
    }

    fn get_provider_list(&self) -> Result<&ProviderList, ToeErrorType> {
        let prov = self.provider_list.get_or_try_init(|| {
            trace!("Lazy initialization of provider list.");
            Ok(load_provider_list(&self.configuration))
        })?;
        match prov {
            Err(e) => Err(e.clone()),
            Ok(p) => Ok(p),
        }
    }
}
