#[derive(Clone, Debug)]
/// All kind of errors AllMyToes may produce.
pub enum ToeErrorType {
    FileDoesNotExist,
    FileNotReadable,
    CannotFetchFileMetadata,
    FilePathNotUTF8Encodable,
    UnsupportedFileType,
    NoCacheIdentifiable,
    XDGConfigDirNotIdentifiable(String),
    FormatCannotBeHandled,
    ErrorOnProbingMimeType,
    ProbingMimeTypeHadNoResult,
    SourceFileCantBeOpened,
    SourceFileCantBeRead,
    TempCantBeCreated(String),
    TempDataCantBeWritten(String),
    TempHeaderCantBeWritten(String),
    UnsupportedImageFormat,
    CacheDoesNotExist(String),
    FileAllocationError(String),
    TempThumbCantBeRenamed(String, String, String),
    ThumbDoesNotExistsAfterCreation,
    ThumbDirCantBeCreated(String),
    CanNotSetPermission(String, String),
    InputFileError(String),
    ThumbLinkingError(String),
    NonPNGFromThumbRepoAsInput,
    CannotOpenPotentialThumbAsPNGForFetchingTEXt(String),
    CannotReadTEXtFromPNG(String),
    InternalError(String),
    NoProviderForMime(String),
    CannotCreateTempDir(String),
    InvalidCommandFormatting(String),
    ProviderSubprocessError(String),
    ProviderHadNoSuccessfulCommand,
    ProviderHadNoSuccessfulMetaCommand(String),
    ProviderCommandDidNotReturnExistingFile(String),
    ImageFromProviderCouldNotBeLoaded(String),
    ProviderSpecFileDoesNotExist(String),
    BadMetaValue(String),
    BadMetaKey(String),
}

impl ToeErrorType {
    /// Returns an human readable error message for the error.
    pub fn msg(&self) -> String {
        match self {
            ToeErrorType::FileDoesNotExist => String::from("File does not exist"),
            ToeErrorType::FileNotReadable => String::from("File is not readable"),
            ToeErrorType::CannotFetchFileMetadata => String::from("Cannot get file-system metadata of input file"),
            ToeErrorType::FilePathNotUTF8Encodable => String::from("The canonical path of the input file cannot be UTF-8 encoded"),
            ToeErrorType::UnsupportedFileType => String::from("Unsupported file type (no regular file or symlink to regular file)"),
            ToeErrorType::NoCacheIdentifiable => String::from("Could not identify a cache dir. At least one of the env vars HOME and XDG_CACHE_HOME must be defined."),
            ToeErrorType::FormatCannotBeHandled => String::from("File format cannot be processed."),
            ToeErrorType::ErrorOnProbingMimeType => String::from("Error on probing MIME type."),
            ToeErrorType::ProbingMimeTypeHadNoResult => String::from("Could not identify MIME type of input file."),
            ToeErrorType::SourceFileCantBeOpened => String::from("Cannot open source file for reading."),
            ToeErrorType::SourceFileCantBeRead => String::from("Error when reading source file."),
            ToeErrorType::TempCantBeCreated(v) => format!("Cannot create temporary thumbnail file {v}"),
            ToeErrorType::TempDataCantBeWritten(v) => format!("Cannot write data chunk to temporary thumbnail file {v}"),
            ToeErrorType::TempHeaderCantBeWritten(v) => format!("Cannot write header to temporary thumbnail file {v}"),
            ToeErrorType::UnsupportedImageFormat => String::from("Unsupported image format"),
            ToeErrorType::CacheDoesNotExist(v) => format!("Cache dir {v} cannot be found on file system."),
            ToeErrorType::FileAllocationError(v) => format!("File allocation error: {v}"),
            ToeErrorType::TempThumbCantBeRenamed(f, t, e) => format!("Temporary thumb file cannot be renamed from {f} to {t}. ({e})"),
            ToeErrorType::ThumbDoesNotExistsAfterCreation => String::from("Thumb has been created but the file does not exist."),
            ToeErrorType::ThumbDirCantBeCreated(v) => format!("Thumbnail directory {v} can't be created."),
            ToeErrorType::CanNotSetPermission(f, m) => format!("Cannot set the permissions for file or directory '{f}'. Error: {m}"),
            ToeErrorType::InputFileError(m) => format!("Unusual error when processing input file: {m}"),
            ToeErrorType::ThumbLinkingError(m) => format!("Error when symlinking thumb. ({m})"),
            ToeErrorType::NonPNGFromThumbRepoAsInput => String::from("A file other than a PNG from within the thumb repository has been given as input."),
            ToeErrorType::CannotOpenPotentialThumbAsPNGForFetchingTEXt(p) => format!("Potential thumb at {p} cannot be opened as PNG."),
            ToeErrorType::CannotReadTEXtFromPNG(p) => format!("Cannot read TEXt meta data from potential thumb at {p}."),
            ToeErrorType::InternalError(s) => format!("Internal error: {s}"),
            ToeErrorType::NoProviderForMime(s) => format!("No provider for mime type {s}"),
            ToeErrorType::InvalidCommandFormatting(s) => format!("Invalid command formatting: {s}"),
            ToeErrorType::ProviderSubprocessError(s) => format!("Error on executing provider command: {s}"),
            ToeErrorType::ProviderHadNoSuccessfulCommand => String::from("No command for the provider was successful."),
            ToeErrorType::ProviderHadNoSuccessfulMetaCommand(s) => format!("No command for the provider's meta key {s} was successful."),
            ToeErrorType::ProviderCommandDidNotReturnExistingFile(s) => format!("Provider command did not place a temporary thumb file at '{s}'"),
            ToeErrorType::ImageFromProviderCouldNotBeLoaded(s) => format!("Provider command did return '{s}', which could not be loaded as file."),
            ToeErrorType::CannotCreateTempDir(s) => format!("Failed to create temporary directory: {s}"),
            ToeErrorType::BadMetaValue(s) => format!("Provider meta-command returned invalid meta data value: {s}"),
            ToeErrorType::BadMetaKey(s) => format!("Provider meta-command returned invalid meta data key: {s}"),
            ToeErrorType::ProviderSpecFileDoesNotExist(s) => format!("Provider specification file {s} does not exist."),
            ToeErrorType::XDGConfigDirNotIdentifiable(s) => format!("Could not identify configuration directory. Neither $XDG_CONFIG_HOME nor $HOME can be read. ({})", s),
        }
    }

    /// Return/exit code for the application per error case
    pub fn error_code(&self) -> i32 {
        match self {
            ToeErrorType::FileDoesNotExist => 1,
            ToeErrorType::FileNotReadable => 2,
            ToeErrorType::CannotFetchFileMetadata => 3,
            ToeErrorType::UnsupportedFileType => 4,
            ToeErrorType::FilePathNotUTF8Encodable => 11,
            ToeErrorType::CacheDoesNotExist(_) => 10,
            ToeErrorType::SourceFileCantBeOpened => 103,
            ToeErrorType::SourceFileCantBeRead => 104,
            ToeErrorType::FormatCannotBeHandled => 103,
            ToeErrorType::UnsupportedImageFormat => 104,
            ToeErrorType::ErrorOnProbingMimeType => 105,
            ToeErrorType::ProbingMimeTypeHadNoResult => 106,
            ToeErrorType::NoCacheIdentifiable => 110,
            ToeErrorType::XDGConfigDirNotIdentifiable(_) => 111,
            ToeErrorType::ThumbDirCantBeCreated(_) => 112,
            ToeErrorType::TempCantBeCreated(_) => 113,
            ToeErrorType::TempDataCantBeWritten(_) => 114,
            ToeErrorType::TempHeaderCantBeWritten(_) => 115,
            ToeErrorType::FileAllocationError(_) => 116,
            ToeErrorType::TempThumbCantBeRenamed(_, _, _) => 117,
            ToeErrorType::ThumbDoesNotExistsAfterCreation => 120,
            ToeErrorType::CanNotSetPermission(_, _) => 121,
            ToeErrorType::InputFileError(_) => 122,
            ToeErrorType::ThumbLinkingError(_) => 123,
            ToeErrorType::NonPNGFromThumbRepoAsInput => 124,
            ToeErrorType::CannotOpenPotentialThumbAsPNGForFetchingTEXt(_) => 130,
            ToeErrorType::CannotReadTEXtFromPNG(_) => 131,
            ToeErrorType::NoProviderForMime(_) => 150,
            ToeErrorType::InvalidCommandFormatting(_) => 151,
            ToeErrorType::ProviderSubprocessError(_) => 152,
            ToeErrorType::ProviderHadNoSuccessfulCommand => 153,
            ToeErrorType::ProviderCommandDidNotReturnExistingFile(_) => 154,
            ToeErrorType::ImageFromProviderCouldNotBeLoaded(_) => 155,
            ToeErrorType::CannotCreateTempDir(_) => 156,
            ToeErrorType::ProviderHadNoSuccessfulMetaCommand(_) => 157,
            ToeErrorType::BadMetaValue(_) => 158,
            ToeErrorType::BadMetaKey(_) => 159,
            ToeErrorType::ProviderSpecFileDoesNotExist(_) => 160,
            ToeErrorType::InternalError(_) => 255,
        }
    }
}
