use dynfmt::{Format, PythonFormat};
use log::{debug, info, log_enabled, trace, warn, Level};
use serde::{self, Deserialize};
use std::collections::HashMap;
use std::path::{Path, PathBuf};
use subprocess::{CaptureData, Exec, Redirection};
use temp_dir::TempDir;

use crate::source::{ReOrientation, Rotation, Source};
use crate::{ThumbSize, ToeErrorType};

pub const DEFAULT_CONFIG_FILE_PATH: &str = include_str!("../conf/provider.yaml");

// #[serde(rename_all = "camelCase")]
#[derive(Debug, PartialEq, Deserialize)]
pub struct Provider {
    mimes: Vec<String>,
    commands: Vec<String>,
    meta: Option<HashMap<String, Vec<String>>>,
    revision: Option<u32>,
}

#[derive(Debug, PartialEq)]
pub struct ProviderList {
    provider: Vec<Provider>,
}

pub struct Provision {
    pub image: image::DynamicImage,
    pub metas: HashMap<String, String>,
}

impl ProviderList {
    pub fn load(path: PathBuf) -> Result<Self, ToeErrorType> {
        let f = std::fs::File::open(path).map_err(|e| {
            ToeErrorType::InternalError(format!("Cannot open provider config file ({})", e))
        })?;
        let default_list_result: Result<Vec<Provider>, serde_yaml::Error> =
            serde_yaml::from_reader(f);
        let default_list = default_list_result.map_err(|e| {
            ToeErrorType::InternalError(format!(
                "Cannot deserialize default provider configuration YAML: {}",
                e
            ))
        })?;

        Ok(ProviderList {
            provider: default_list,
        })
    }

    pub fn default() -> Result<Self, ToeErrorType> {
        let default_list_result: Result<Vec<Provider>, serde_yaml::Error> =
            serde_yaml::from_str(DEFAULT_CONFIG_FILE_PATH);

        let default_list = default_list_result.map_err(|e| {
            ToeErrorType::InternalError(format!(
                "Cannot deserialize default provider configuration YAML: {}",
                e
            ))
        })?;

        Ok(ProviderList {
            provider: default_list,
        })
    }

    pub fn find_provider(&self, mime: &str) -> Option<&Provider> {
        trace!("Looking for a provider for {}...", mime);
        for p in &self.provider {
            for m in &p.mimes {
                if m == mime {
                    return Some(p);
                }
            }
        }
        None
    }
}

impl Provider {
    fn invoke_until_success(
        command_list: &Vec<String>,
        args: &HashMap<String, &String>,
    ) -> Result<Option<CaptureData>, ToeErrorType> {
        for command_str_raw in command_list {
            let command_str = PythonFormat
                .format(command_str_raw, args)
                .map_err(|e| ToeErrorType::InvalidCommandFormatting(format!("{}", e)))?;
            debug!("Provider command call: {}", command_str.as_ref());
            match Exec::shell(command_str.as_ref())
                .stdout(Redirection::Pipe)
                .stderr(Redirection::Pipe)
                .capture()
            {
                Err(e) => {
                    info!(
                        "Provider command '{}' failed ({}). Trying next.",
                        command_str.as_ref(),
                        e
                    );
                }
                Ok(call_result) => {
                    if log_enabled!(Level::Debug) {
                        let stdout = call_result.stdout_str();
                        let stderr = call_result.stderr_str();
                        if !stdout.is_empty() {
                            debug!("Command stdout: {}", stdout);
                        }
                        if !stderr.is_empty() {
                            debug!("Command stderr: {}", stderr);
                        }
                    }
                    if call_result.exit_status == subprocess::ExitStatus::Exited(0) {
                        debug!("Provider command executed successful.");
                        return Ok(Some(call_result));
                    } else {
                        warn!(
                            "Provider command executed with {:?}. Trying next.",
                            call_result.exit_status
                        );
                    }
                }
            }
        }
        Ok(None)
    }

    pub fn get_results(
        &self,
        path: &String,
        size: &ThumbSize,
        leak: bool,
    ) -> Result<Provision, ToeErrorType> {
        let size_str = size.max_thumb_width().to_string();
        let temp_dir =
            TempDir::new().map_err(|e| ToeErrorType::CannotCreateTempDir(format!("{}", e)))?;
        let temp_dir_string = temp_dir.path().as_os_str().to_string_lossy().into_owned();
        let temp_file_path = temp_dir.child("thumbimage");
        let outfile_string = temp_file_path.as_os_str().to_string_lossy().into_owned();
        let args = HashMap::from([
            ("file".to_string(), path),
            ("size".to_string(), &size_str),
            ("outfile".to_string(), &outfile_string),
            ("tmpdir".to_string(), &temp_dir_string),
        ]);

        // Call command to create thumb image
        let command_result = Provider::invoke_until_success(&self.commands, &args)?;

        // Call commands to create meta data
        let mut metas = HashMap::new();
        if let Some(revision) = self.revision {
            metas.insert("AMT::ProviderRevision".to_string(), revision.to_string());
        }
        if let Some(meta) = &self.meta {
            for (meta_key, meta_commands) in meta.iter() {
                if meta_key.is_empty() {
                    return Err(ToeErrorType::BadMetaKey("Length is zero.".to_string()));
                }
                if meta_key.len() > 79 {
                    return Err(ToeErrorType::BadMetaKey(
                        "Lenght is greater than 79.".to_string(),
                    ));
                }
                let meta_result = Provider::invoke_until_success(meta_commands, &args)?.ok_or(
                    ToeErrorType::ProviderHadNoSuccessfulMetaCommand(meta_key.to_string()),
                )?;
                let stdout = meta_result.stdout_str().trim().to_string();
                debug!("Add meta data: {} → {}", meta_key, stdout);
                if stdout.is_empty() {
                    return Err(ToeErrorType::BadMetaValue("Stdout is empty.".to_string()));
                }
                if stdout.len() > 128 {
                    return Err(ToeErrorType::BadMetaValue(
                        "Stdout is longer than 128.".to_string(),
                    ));
                }
                if stdout.contains('\n') || stdout.contains('\r') {
                    return Err(ToeErrorType::BadMetaValue(
                        "Stdout contains newline or carriage return.".to_string(),
                    ));
                }
                metas.insert(meta_key.to_string(), stdout);
            }
        }

        // Leak the temporary image if we've been asked to do so
        if leak {
            info!("Leaking temporary directory {:?}", temp_dir.path());
            temp_dir.leak();
        }

        // Load image and return
        match command_result {
            None => Err(ToeErrorType::ProviderHadNoSuccessfulCommand),
            Some(_) => {
                if Path::new(&temp_file_path).is_file() {
                    debug!("Loading image file that has been provided by provider command...");
                    let source = Source::load(&outfile_string)?;
                    if let Some(image) = source.image {
                        if let Some(reor) = source.reorientation {
                            if reor
                                != (ReOrientation {
                                    rotation: Rotation::R0,
                                    flipped: false,
                                })
                            {
                                warn!("Provider returned image if EXIF reorientation. This will be ignored. The thumb might be not be as expected.")
                            }
                        }
                        Ok(Provision { image, metas })
                    } else {
                        Err(ToeErrorType::ImageFromProviderCouldNotBeLoaded(
                            outfile_string,
                        ))
                    }
                } else {
                    Err(ToeErrorType::ProviderCommandDidNotReturnExistingFile(
                        outfile_string,
                    ))
                }
            }
        }
    }
}
