use image::DynamicImage;
use log::{debug, error, info, trace};
use std::fs::File;
use std::path::{Path, PathBuf};

use crate::errors::ToeErrorType;
use crate::source::Source;
use crate::source::{ReOrientation, Rotation};
use crate::sourcedescriptor::{read_meta, SourceDescriptor};
use crate::thumb::Thumb;
use crate::thumbsize::ThumbSize;
use crate::writer;
use crate::AMT;

#[derive(Debug)]
pub struct ThumbRequest<'a> {
    pub path: &'a Path,
    pub size: ThumbSize,
    pub amt: &'a AMT,
}

impl ThumbRequest<'_> {
    pub fn get(&self) -> Result<Thumb, ToeErrorType> {
        // Calculate and gather some basic data for the job the scenario
        let source_descriptor = SourceDescriptor::new(self.path)?;
        trace!("{:?}", source_descriptor);

        // Only proceed if the input file is either a regular file or a symlink to a regular file
        if !source_descriptor.metadata.is_file() {
            return Err(ToeErrorType::UnsupportedFileType);
        }

        // Only proceed if the input file is readable
        assess_input_file_is_readable(&source_descriptor)?;

        // Early termination if file is from within the thumb repository
        if let Some(thumb) = result_for_input_from_within_thumb_repo(&source_descriptor)? {
            return Ok(thumb);
        }

        // Check for existing thumb and return on success
        if !self.amt.configuration.force_creation {
            if let Some(thumb) = source_descriptor.get_thumb_if_existing_and_valid(self.size)? {
                return Ok(thumb);
            }
        }

        // Calculate feasible size
        let (feasible_size, input_dimensions) =
            self.calc_feasible_size(&source_descriptor.file_path);

        // If a thumb exists for the feasible size, return that one
        if let Some(some_feasible_size) = feasible_size {
            if some_feasible_size != self.size {
                assert!(some_feasible_size <= self.size);
                if let Some(thumb) =
                    source_descriptor.get_thumb_if_existing_and_valid(some_feasible_size)?
                {
                    info!(
                        "Feasible smaller thumb size ({:?}) with exisiting thumb found.",
                        feasible_size
                    );
                    return Ok(thumb);
                } else {
                    debug!("Feasible smaller thumb size found ({:?}), but also that one does not have a thumbnail yet.", feasible_size);
                }
            } else {
                debug!("Feasible size equals the requested size.");
            }
        }

        // Create Thumbs
        let result_size =
            self.create_thumbnails(&source_descriptor, feasible_size, input_dimensions)?;

        // Do a final check that the new thumbnail does exist now
        if let Some(result_thumb) =
            source_descriptor.get_thumb_if_existing_and_valid(result_size)?
        {
            trace!("Thumb created. Returning it now.");
            Ok(result_thumb)
        } else {
            trace!("Thumb should have been created but it doesn't seem to exists.");
            Err(ToeErrorType::ThumbDoesNotExistsAfterCreation)
        }
    }

    fn calc_feasible_size(
        &self,
        input_image_path: &String,
    ) -> (Option<ThumbSize>, Option<(u32, u32)>) {
        // Check if we can pick a feasible size in case the input is smaller than the thumb and if we can
        // determine the image's dimensions without loading it.
        if self.amt.configuration.return_smallest_feasible {
            match imagesize::size(input_image_path) {
                Ok(input_dimensions) => {
                    let id_w = input_dimensions.width as u32;
                    let id_h = input_dimensions.height as u32;
                    debug!(
                        "Early loaded dimensions of input image: {} x {}",
                        id_w, id_h
                    );
                    let original_max_edge_length = std::cmp::max(id_w, id_h);
                    if original_max_edge_length < self.size.max_thumb_width() {
                        let target_size =
                            ThumbSize::smallest_greater_or_equal(original_max_edge_length);
                        match target_size {
                            None => {
                                debug!(
                                    "No thumb size is bigger or equal to the size of the input image."
                                );
                                (Some(self.size), Some((id_w, id_h)))
                            }
                            Some(feasible_size) => (Some(feasible_size), Some((id_w, id_h))),
                        }
                    } else {
                        (Some(self.size), Some((id_w, id_h)))
                    }
                }
                Err(why) => {
                    debug!("Error getting dimensions of input file ({:?}). Image needs to be loaded first.", why);
                    (None, None)
                }
            }
        } else {
            (None, None)
        }
    }

    fn create_thumbnails(
        &self,
        source_descriptor: &SourceDescriptor,
        feasible_size: Option<ThumbSize>,
        input_dimensions: Option<(u32, u32)>,
    ) -> Result<ThumbSize, ToeErrorType> {
        let mut extra_meta_tuples: Vec<(String, String)> = Vec::new();

        let source = Source::load(&source_descriptor.file_path)?;

        extra_meta_tuples.extend(vec![
            (
                String::from("Thumb::Size"),
                source_descriptor.metadata.len().to_string(),
            ),
            (String::from("Thumb::Mimetype"), format!("{}", source.mime)),
        ]);

        match source.image {
            Some(source_image) => self.create_thumbnails_from_image(
                source_descriptor,
                feasible_size,
                input_dimensions,
                source_image,
                &source.reorientation,
                extra_meta_tuples,
            ),
            None => self.create_thumbnails_from_provider(
                source_descriptor,
                source.mime,
                extra_meta_tuples,
            ),
        }
    }

    fn create_thumbnails_from_provider(
        &self,
        source_descriptor: &SourceDescriptor,
        mime: mime::Mime,
        mut extra_meta_tuples: Vec<(String, String)>,
    ) -> Result<ThumbSize, ToeErrorType> {
        let leak = self.amt.configuration.leak;
        let provider_list = self.amt.get_provider_list()?;
        if let Some(provider) = provider_list.find_provider(mime.as_ref()) {
            debug!("Found provider for mime {}.", mime);
            let provision = provider.get_results(&source_descriptor.file_path, &self.size, leak)?;
            let sizes_for_new_thumbs = vec![self.size];
            let mut source_image = provision.image;
            let reorientation = None;
            let original_max_edge_length = self.size.max_thumb_width();
            extra_meta_tuples.extend(provision.metas);

            if source_image.color() == image::ColorType::Rgba8 {
                trace!("Image from provider is 8-bit RGBA. No conversion needed.");
            } else {
                trace!("Image from provider is not 8-bit RGBA and needs to be converted.");
                source_image = image::DynamicImage::ImageRgba8(source_image.into_rgba8());
            };

            self.generate_thumbs_for_sizes(
                source_descriptor,
                source_image,
                &reorientation,
                original_max_edge_length,
                &sizes_for_new_thumbs,
                extra_meta_tuples,
            )?;

            Ok(self.size)
        } else {
            trace!("Did not find a provider for mime {}.", mime);
            Err(ToeErrorType::NoProviderForMime(mime.to_string()))
        }
    }

    fn create_thumbnails_from_image(
        &self,
        source_descriptor: &SourceDescriptor,
        mut feasible_size: Option<ThumbSize>,
        input_dimensions: Option<(u32, u32)>,
        mut source_image: DynamicImage,
        reorientation: &Option<ReOrientation>,
        mut extra_meta_tuples: Vec<(String, String)>,
    ) -> Result<ThumbSize, ToeErrorType> {
        let original_size = (source_image.width(), source_image.height());
        let original_max_edge_length = std::cmp::max(original_size.0, original_size.1);
        trace!("Image loaded with size {:?}", original_size);

        if source_image.color() == image::ColorType::Rgba8 {
            trace!("Image is 8-bit RGBA. No conversion needed.");
        } else {
            trace!("Image is not 8-bit RGBA and needs to be converted.");
            source_image = image::DynamicImage::ImageRgba8(source_image.into_rgba8());
        };

        if let Some(precalculated_dimension) = input_dimensions {
            if precalculated_dimension != original_size {
                error!("Precalculated size and size of loaded image are different! Ignoring any precalcualted, feasible thumb size.");
                feasible_size = None;
            }
        }

        let target_size = if let Some(s) = feasible_size {
            s
        } else if !self.amt.configuration.return_smallest_feasible
            || original_max_edge_length >= self.size.max_thumb_width()
        {
            self.size
        } else {
            ThumbSize::smallest_greater_or_equal(original_max_edge_length).unwrap_or(self.size)
        };
        trace!("Going for thumb size {:?}", target_size);

        let sizes_for_new_thumbs = vec![target_size];

        extra_meta_tuples.extend(vec![
            (
                String::from("Thumb::Image::Width"),
                original_size.0.to_string(),
            ),
            (
                String::from("Thumb::Image::Height"),
                original_size.1.to_string(),
            ),
            (
                String::from("Thumb::Image::Dimensions"),
                format!("{}x{}", original_size.0, original_size.1),
            ),
        ]);

        if let Some(reor) = reorientation {
            extra_meta_tuples.push((
                String::from("Thumb::Image::Reorientation"),
                reor.to_string(),
            ))
        };

        self.generate_thumbs_for_sizes(
            source_descriptor,
            source_image,
            reorientation,
            original_max_edge_length,
            &sizes_for_new_thumbs,
            extra_meta_tuples,
        )?;

        Ok(target_size)
    }

    fn generate_thumbs_for_sizes(
        &self,
        request: &SourceDescriptor,
        mut source_content: DynamicImage,
        reorientation: &Option<ReOrientation>,
        original_max_edge_length: u32,
        sizes: &[ThumbSize],
        extra_meta_tuples: Vec<(String, String)>,
    ) -> Result<(), ToeErrorType> {
        for (ix, &size) in sizes.iter().enumerate() {
            if self.amt.configuration.force_creation
                || request.get_thumb_if_existing_and_valid(size)?.is_none()
            {
                trace!("Thumb with size {} will be created.", size.dir_name());
                let target_path = &request.get_thumb_path_for_size(size)?;
                let target_size = if original_max_edge_length > size.max_thumb_width() {
                    size.max_thumb_width()
                } else {
                    original_max_edge_length
                };
                // For performance reasons, re-orientation is performed only on the biggest thumb.
                // Smaller thumbs are created from the biggest thumb.
                // Performing re-orientation on the original is more expensive because the original is
                // usually bigger than the biggest thumb.
                source_content = generate_one_thumbnail(
                    request,
                    &source_content,
                    target_path,
                    target_size,
                    if ix == 0 { reorientation } else { &None },
                    &extra_meta_tuples,
                )?;
            } else {
                trace!("Thumb of size {} does exist. Skipping.", size.dir_name());
            }
        }
        Ok(())
    }
}

/// Checks if a file is readable.
/// If so, `Ok()` is returned. Otherwise a `Err(ToeErrorType::FileNotReadable)`.
fn assess_input_file_is_readable(request: &SourceDescriptor) -> Result<(), ToeErrorType> {
    if File::open(&request.file_path).is_err() {
        Err(ToeErrorType::FileNotReadable)
    } else {
        Ok(())
    }
}

/// Check if a file is located inside of the thumb cache directory.
///
/// Returns `None` if the input file is not from the thumb cache directory.
/// That's the normal case.
///
/// Returns Some(path) to the input file itself in case it looks like a PNG image.
/// That's a work-around to provide clients some PNG as "thumb" for a thumb used as input.
///
/// Returns Some(Err) in case the input file is from within the thumb cache but is not a PNG file.
fn result_for_input_from_within_thumb_repo(
    request: &SourceDescriptor,
) -> Result<Option<Thumb>, ToeErrorType> {
    let input_file_path = PathBuf::from(&request.file_path);
    if input_file_path.starts_with(&request.thumb_dir) {
        trace!("Input file is located within the thumbnail repository.");
        if let Some(extension) = input_file_path.extension() {
            if extension == "png" {
                trace!("Input file has extension 'png'. Returning the input file as thumb.");
                let meta = read_meta(&input_file_path)?;
                return Ok(Some(Thumb {
                    path: String::from(&request.file_path),
                    hash: String::from(&request.uri_hash),
                    meta,
                }));
            }
        }
        trace!(
            "Input file {:?} is from within the thumbnail repository but not a PNG. Terminating.",
            input_file_path
        );
        return Err(ToeErrorType::NonPNGFromThumbRepoAsInput);
    }
    Ok(None)
}

/// Generates one thumbnail (for one size), based on an input image and a given target-size.
///
/// The scaled thumbnail is also returned for reuse in another, smaller, thumbnail-generation
/// iteration.
///
/// This function basically only does the scaling and uses `writer::save_thumb` to actually store
/// the thumbnail with its meta data.
fn generate_one_thumbnail(
    toe: &SourceDescriptor,
    source_image: &image::DynamicImage,
    dest_path: &std::path::Path,
    size: u32,
    reorientation: &Option<ReOrientation>,
    extra_meta_tuples: &Vec<(String, String)>,
) -> Result<image::DynamicImage, ToeErrorType> {
    trace!("Creating thumb with max. edge of {}px.", size,);

    let mut thumb_image = source_image.thumbnail(size, size);
    if let Some(reor) = reorientation {
        debug!("Apply re-orientation based on EXIF data {:?}.", reor);
        match reor.rotation {
            Rotation::R0 => {}
            Rotation::R90 => thumb_image = thumb_image.rotate270(),
            Rotation::R180 => thumb_image = thumb_image.rotate180(),
            Rotation::R270 => thumb_image = thumb_image.rotate90(),
        }
        if reor.flipped {
            thumb_image = thumb_image.fliph();
        }
    }

    writer::save_thumb(toe, &thumb_image, dest_path, extra_meta_tuples)?;

    debug!("Created '{:?}'", dest_path.as_os_str());
    Ok(thumb_image)
}
