#[derive(PartialOrd, Ord, PartialEq, Eq, Clone, Copy, Debug)]
/// The available sizes for thumbnails.
pub enum ThumbSize {
    Normal,
    Large,
    XLarge,
    XXLarge,
}

impl ThumbSize {
    /// Static function which returns all available thumb sizes in order from biggest to smallest.
    pub fn order() -> [ThumbSize; 4] {
        [
            ThumbSize::Normal,
            ThumbSize::Large,
            ThumbSize::XLarge,
            ThumbSize::XXLarge,
        ]
    }

    /// Returns the edge length in pixels.
    pub fn max_thumb_width(&self) -> u32 {
        match self {
            ThumbSize::Normal => 128,
            ThumbSize::Large => 256,
            ThumbSize::XLarge => 512,
            ThumbSize::XXLarge => 1024,
        }
    }

    /// Returns the size-specific sub-directory name.
    pub fn dir_name(&self) -> &str {
        match self {
            ThumbSize::Normal => "normal",
            ThumbSize::Large => "large",
            ThumbSize::XLarge => "x-large",
            ThumbSize::XXLarge => "xx-large",
        }
    }

    /// Returns the smallest thumb size which has a max-width bigger or equal to the argument.
    /// If the argument is bigger than any thumb size, `None` is returned.
    pub fn smallest_greater_or_equal(width: u32) -> Option<ThumbSize> {
        ThumbSize::order()
            .into_iter()
            .find(|&ts| ts.max_thumb_width() >= width)
    }
}

#[cfg(test)]
mod test_thumb_size_order {
    use super::ThumbSize;
    #[test]
    fn order() {
        for win in ThumbSize::order().windows(2) {
            assert_eq!(2, win.len());
            println!("Window: {:?}/{:?}", win[0], win[1]);
            assert!(win[0] < win[1]);
            assert!(win[0].max_thumb_width() < win[1].max_thumb_width());
        }
    }
}

#[cfg(test)]
mod test_thumb_size_smallest_greater_or_equal {
    use super::ThumbSize;

    #[test]
    fn smaller_than_normal() {
        assert_eq!(
            Some(ThumbSize::Normal),
            ThumbSize::smallest_greater_or_equal(100)
        );
    }

    #[test]
    fn smaller_than_xxlarge() {
        assert_eq!(
            Some(ThumbSize::XXLarge),
            ThumbSize::smallest_greater_or_equal(1000)
        );
    }

    #[test]
    fn greater_than_all() {
        assert_eq!(None, ThumbSize::smallest_greater_or_equal(1050));
    }
}
